/*  sha256.c

    Copyright (C) 2008-2015 bla <blapost@gmail.com>
    all rights reserved.

    sha256 as defined by fips180-2.pdf
*/

#include <stdint.h>
#include <string.h>


static uint32_t K[] = {
        0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5,
        0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
        0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
        0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
        0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc,
        0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
        0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7,
        0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
        0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
        0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
        0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3,
        0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
        0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5,
        0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
        0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
        0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
        };
static uint32_t H_init[8] = {
        0x5be0cd19, 0x1f83d9ab, 0x9b05688c, 0x510e527f,
        0xa54ff53a, 0x3c6ef372, 0xbb67ae85, 0x6a09e667
        };
static uint32_t Ch(uint32_t x, uint32_t y, uint32_t z)
{
    return (x & y) ^ (~x & z);
}

static uint32_t Maj(uint32_t x, uint32_t y, uint32_t z)
{
    return (x & y) ^ (x & z) ^ (y & z);
}

static uint32_t ROTR(uint32_t x, uint32_t pos)
{
    pos &= 31;
    return (x >> pos) | (x << (32 - pos));
}

static uint32_t sum0(uint32_t x)
{
    return ROTR(x, 2) ^ ROTR(x, 13) ^ ROTR(x, 22);
}

static uint32_t sum1(uint32_t x)
{
    return ROTR(x, 6) ^ ROTR(x, 11) ^ ROTR(x, 25);
}

uint32_t sigma0(uint32_t x)
{
    return ROTR(x, 7) ^ ROTR(x, 18) ^ (x >> 3);
}

static uint32_t sigma1(uint32_t x)
{
    return ROTR(x, 17) ^ ROTR(x, 19) ^ (x >> 10);
}

static uint32_t H[8];
static void processchunk(unsigned char *data){
    uint32_t W[64];
    uint32_t a = H[7], b = H[6], c = H[5], d = H[4];
    uint32_t e = H[3], f = H[2], g = H[1], h = H[0];
    uint32_t t1, t2;
    int t,i;

    for(t = 0; t < 16; ++t)
        for(i = 0; i < 4; ++i)
        W[t] = W[t] << 8 | data[t*4 + i];


    for(t = 16; t < 64; ++t)
        W[t] = sigma1(W[t-2]) + W[t - 7] + sigma0(W[t - 15]) + W[t - 16];


    for(t = 0; t < 64; ++t)
    {
        t1 = h + sum1(e) + Ch(e, f, g) + K[t] + W[t];
        t2 = sum0(a) + Maj(a, b, c);
        h = g;
        g = f;
        f = e;
        e = d + t1;
        d = c;
        c = b;
        b = a;
        a = t1 + t2;
    }

    H[7] += a;
    H[6] += b;
    H[5] += c;
    H[4] += d;
    H[3] += e;
    H[2] += f;
    H[1] += g;
    H[0] += h;
}

int checksha256(unsigned char *data, unsigned int size,unsigned char* targethash){
    uint32_t len = size * 8, i;
    data[size++] = 0x80;
    while(size % 64 != 60)
        data[size++] = 0;

    for(i = 0; i < 4; i++)
        data[size++] = len = ROTR(len, 24);

    memcpy(H, H_init, sizeof(H_init));
    for(i = 0; i < size; i += 64){
        processchunk(data + i);
    }

    return !memcmp(H, targethash, 32);
}


